import os
import shutil

import pytest

import genomepy.utils
from tests import linux


def test_mkdir_p(path="./tests/dir1/dir2/nestled_dir"):
    shutil.rmtree("./tests/dir1", ignore_errors=True)
    assert not os.path.isdir(path)

    # make a non-existing nestled-dir
    genomepy.utils.mkdir_p(path)
    assert os.path.isdir(path)

    # try to make an existing dir
    genomepy.utils.mkdir_p(path)

    shutil.rmtree("./tests/dir1", ignore_errors=True)
    assert not os.path.isdir(path)


def test_rm_rf(path="./tests/dir1/dir2/nestled_dir"):
    genomepy.utils.mkdir_p(path)
    fname = os.path.join(path, "file.txt")
    with open(fname, "w") as f:
        f.write("hello world!")

    # try to remove an existing file
    assert os.path.isfile(fname)
    genomepy.utils.rm_rf(fname)
    assert not os.path.isfile(fname)

    # try to remove an existing dir
    assert os.path.isdir(path)
    genomepy.utils.rm_rf(path)
    assert not os.path.isdir(path)

    # try to remove a non-existing dir
    genomepy.utils.rm_rf(path)


def test_get_genomes_dir(genomes_dir="tests/data"):
    # dir does exist
    gd = genomepy.utils.get_genomes_dir(genomes_dir=genomes_dir, check_exist=True)
    assert os.path.abspath(gd) == os.path.abspath(genomes_dir)

    # dir does not exist
    with pytest.raises(FileNotFoundError):
        genomepy.utils.get_genomes_dir(genomes_dir="fake_dir", check_exist=True)


def test_cmd_ok():
    assert genomepy.utils.cmd_ok("STAR")
    assert genomepy.utils.cmd_ok("bwa")
    assert not genomepy.utils.cmd_ok("missing_cmd")


def test_run_index_cmd(caplog, name="tests", good_cmd="ls", bad_cmd="bad_cmd"):
    genomepy.utils.run_index_cmd(name=name, cmd=good_cmd)

    # bad_command not found error
    genomepy.utils.run_index_cmd(name=name, cmd=bad_cmd)
    # captured = capsys.readouterr().err.strip()
    if linux:
        assert f"{bad_cmd}: not found" in caplog.text
    else:
        # thanks for being different mac. That took me 30 minutes of my life...
        assert f"{bad_cmd}: command not found" in caplog.text


def test_safe(unsafe_name="a name ", safe_name="a_name"):
    result = genomepy.utils.safe(unsafe_name)
    assert result == safe_name


def test_lower(unsafe_name="A nAme ", safe_name="a_name"):
    result = genomepy.utils.lower(unsafe_name)
    assert result == safe_name


def test_get_localname(name="XENTR_9.1", localname="my genome"):
    # localname input
    result = genomepy.utils.get_localname(name=name, localname=localname)
    assert result == genomepy.utils.safe(localname)

    # name input
    result = genomepy.utils.get_localname(name=name)
    assert result == genomepy.utils.safe(name)

    # URL input (simple)
    url = "http://ftp.xenbase.org/pub/Genomics/JGI/Xentr9.1/XENTR_9.1_genome.fa.gz"
    result = genomepy.utils.get_localname(name=url)
    assert result == name

    # URL input (complex)
    url2 = (
        "https://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/027/325/"
        "GCF_000027325.1_ASM2732v1/GCF_000027325.1_ASM2732v1_genomic.fna.gz"
    )
    result = genomepy.utils.get_localname(name=url2)
    assert result == "GCF_000027325.1_ASM2732v1"

    # Local path input
    path = "tests/data/sacCer3/sacCer3.fa"
    result = genomepy.utils.get_localname(name=path)
    assert result == "sacCer3"


def test_try_except_pass():
    def raise_error():
        raise ValueError

    # expected error is caught
    genomepy.utils.try_except_pass(ValueError, raise_error)
    genomepy.utils.try_except_pass((ValueError, FileNotFoundError), raise_error)

    # unexpected error is not caught
    with pytest.raises(ValueError):
        genomepy.utils.try_except_pass(FileNotFoundError, raise_error)


def test_check_ucsc_tools():
    genomepy.utils.check_ucsc_tools()

    with pytest.raises(OSError):
        genomepy.utils.check_ucsc_tools(["this_tool_does_not_exist_1234567890"])



import os
from time import sleep

from loguru import logger

from genomepy.caching import cache_exp_genomes, disk_cache, lock
from genomepy.exceptions import GenomeDownloadError
from genomepy.files import update_readme
from genomepy.online import check_url, connect_ftp_link
from genomepy.providers.base import BaseProvider, download_annotation
from genomepy.providers.ncbi import download_assembly_report, get_genome_size
from genomepy.providers.ucsc import UcscProvider
from genomepy.utils import get_genomes_dir, get_localname


class GencodeProvider(BaseProvider):
    """
    GENCODE genome provider.

    GENCODE sports superb annotations for human and mouse with UCSC-style chromosome names.
    Genomes on this provider are unmasked, so we use the UCSC genomes instead.

    Note: this combination lacks scaffolds and alternate haplotype sequences
    (their names don't match between GENCODE and UCSC).
    """

    name = "GENCODE"
    accession_fields = ["assembly_accession"]
    taxid_fields = ["taxonomy_id"]
    description_fields = [
        "species",
        "other_info",
        "text_search",
    ]
    _cli_install_options = {}
    _url = "ftp://ftp.ebi.ac.uk/pub/databases/gencode"

    def __init__(self):
        self._provider_status()
        # Populate on init, so that methods can be cached
        self.genomes = _get_genomes(self._url)
        self.ucsc = UcscProvider()
        self.gencode2ucsc = get_gencode2ucsc(self.genomes)
        self._update_genomes()

    @staticmethod
    def ping():
        """Can the provider be reached?"""
        ftp_online = bool(check_url("ftp.ebi.ac.uk/pub/databases/gencode"))
        return ftp_online

    def _genome_info_tuple(self, name, size=False):
        """tuple with assembly metadata"""
        accession = self.genomes[name]["assembly_accession"]
        taxid = self.genomes[name]["taxonomy_id"]
        annotations = True
        species = self.genomes[name]["species"]
        other = self.genomes[name]["other_info"]
        if size:
            length = get_genome_size(accession)
            return name, accession, taxid, annotations, species, length, other
        return name, accession, taxid, annotations, species, other

    def _update_genomes(self):
        """add assembly accession and other information to the genomes dict"""
        for name in self.genomes:
            self.genomes[name]["other_info"] = "GENCODE annotation + UCSC genome"
            # makes the genome findable with UCSC names
            ucsc_name = self.gencode2ucsc[name]
            self.genomes[name]["text_search"] += f" {ucsc_name}"
            # add the UCSC accession ID
            ucsc_acc = self.ucsc.genomes[ucsc_name]["assembly_accession"]
            self.genomes[name]["assembly_accession"] = ucsc_acc

    def get_genome_download_link(self, name: str, mask: str = "soft", **kwargs) -> str:
        """
        Return UCSC http link to genome sequence

        Parameters
        ----------
        name : str
            Genome name.

        mask : str , optional
            Masking level. Options: soft, hard or none. Default is soft.

        Returns
        ------
        str
            http/ftp link tp genome.
        """
        ucsc_name = self.gencode2ucsc[name]
        return self.ucsc.get_genome_download_link(ucsc_name, mask, **kwargs)

    def download_genome(
        self,
        name: str,
        genomes_dir: str = None,
        localname: str = None,
        mask: str = "soft",
        **kwargs,
    ):
        """
        Download genomes from UCSC, as the GENCODE genomes aren't masked.
        Contigs between the UCSC genome and GENCODE annotations match.
        """
        ucsc_name = self.gencode2ucsc[name]
        self.ucsc.name = "GENCODE"  # for logging & readme
        self.ucsc.download_genome(ucsc_name, genomes_dir, localname, mask, **kwargs)
        self.ucsc.name = "UCSC"

    def get_annotation_download_links(self, name, **kwargs):
        """
        Retrieve functioning gene annotation download link(s).

        Parameters
        ----------
        name : str
            genome name

        Returns
        -------
        list
            http/ftp link(s)
        """
        return self.genomes[name]["annotations"]

    def download_annotation(self, name, genomes_dir=None, localname=None, **kwargs):
        """
        Download annotation file to to a specific directory

        Parameters
        ----------
        name : str
            Genome / species name

        genomes_dir : str , optional
            Directory to install annotation

        localname : str , optional
            Custom name for your genome
        """
        name = self._check_name(name)
        link = self.get_annotation_download_link(name, **kwargs)

        localname = get_localname(name, localname)
        genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)

        logger.info(f"Downloading annotation from {self.name}. Target URL: {link}...")
        try:
            # download exact assembly report to rename the scaffolds
            acc = self.assembly_accession(name)
            fname = os.path.join(genomes_dir, localname, "assembly_report.txt")
            download_assembly_report(acc, fname)

            download_annotation(genomes_dir, link, localname)
            logger.info("Annotation download successful")
        except Exception as e:
            raise GenomeDownloadError(
                f"An error occured while installing the gene annotation for {name} from {self.name}.\n"
                "If you think the annotation should be there, please file a bug report at: "
                "https://github.com/vanheeringen-lab/genomepy/issues\n\n"
                f"Error: {e.args[0]}"
            )

        # Add annotation URL to readme
        readme = os.path.join(genomes_dir, localname, "README.txt")
        update_readme(readme, updated_metadata={"annotation url": link})


def get_gencode2ucsc(genomes):
    """
    generate conversion dict for gencode's
    Ensembl-style names to UCSC-style names.

    Parameters
    ----------
    genomes : dict
        the provider's genomes dict

    Returns
    -------
    dict
        keys: Ensembl names, values: UCSC names
    """
    # start with historic assemblies
    gencode2ucsc = {
        "GRCh37": "hg19",
        "GRCm38": "mm10",
    }
    # add latest assemblies
    for assembly in genomes:
        if assembly not in gencode2ucsc:
            specie = genomes[assembly]["species"]
            specie = "hg" if specie.startswith("H") else "mm"
            number = "".join([s for s in assembly if s.isdigit()])
            ucsc_assembly = f"{specie}{number}"
            gencode2ucsc[assembly] = ucsc_assembly
    return gencode2ucsc


def get_releases(listing: list, specie: str) -> list:
    """
    return the list of gencode releases found on the site.
    Highest fist.

    Parameters
    ----------
    listing : list
        list of gencode FTP paths, some ending with "release_nn" or "release_Mnn"

    specie : str
        mouse of human. mouse release numbers are prefixed with "M"

    Returns
    ------
    list
        releases sorted from highest to lowest
    """
    # ignore releases without a consistent system.
    releases = [int(ftpdir[-2:]) for ftpdir in listing if ftpdir[-2:].isdigit()]
    releases = sorted([str(n) for n in releases if n > 21], reverse=True)
    if specie == "mouse":
        releases = ["M" + r for r in releases]  # dear gencode: why?
    return releases


def add_grch37(genomes, ftp_link):
    """gencode's GRCh37 is filed with a unique system"""
    # no mouse in genomes yet
    latest_human = f"GRCh{max([int(g[-2:]) for g in genomes])}"
    latest_annot = genomes[latest_human]["annotations"][0]
    release = [r for r in latest_annot.split("/") if "release_" in r][0][-2:]
    genomes["GRCh37"] = {
        "annotations": [
            f"{ftp_link}/Gencode_human/release_{release}/GRCh37_mapping/"
            f"gencode.v{release}lift37.annotation.gtf.gz"
        ],
        "taxonomy_id": 9606,
        "species": "Homo sapiens",
        "text_search": "human",
    }
    return genomes


def _get_genomes(ftp_link):
    """Wrapper to catch FTP issues."""
    try:
        genomes = get_genomes(ftp_link)
    except (TimeoutError, ConnectionRefusedError):
        logger.warning(
            "GENCODE is online, but cannot be reached. "
            "Is FTP working on this device?"
        )
        genomes = {}
    return genomes


@lock
@disk_cache.memoize(expire=cache_exp_genomes, tag="get_genomes-gencode")
def get_genomes(ftp_link):
    """genomes dict of the latest gencode release of each major assembly."""
    logger.info("Downloading assembly summaries from GENCODE")

    genomes = {}
    species = {"human": "Homo sapiens", "mouse": "Mus musculus"}
    taxid = {"human": 9606, "mouse": 10090}
    sleep(1)  # we just closed a connection with ping()
    ftp, ftp_path = connect_ftp_link(ftp_link, timeout=7)
    for specie in ["human", "mouse"]:
        listing = ftp.nlst(f"{ftp_path}/Gencode_{specie}")
        releases = get_releases(listing, specie)
        for release in releases:
            listing = ftp.nlst(f"{ftp_path}/Gencode_{specie}/release_{release}")
            files = [f.split("/")[-1] for f in listing]
            # drop patch level (UCSC genomes don't have patches either)
            assembly = [f for f in files if "primary_assembly" in f][0].split(".")[0]
            if assembly not in genomes:
                genomes[assembly] = {
                    "annotations": [
                        f"{ftp_link}/Gencode_{specie}/release_{release}/"
                        f"gencode.v{release}.primary_assembly.annotation.gtf.gz"
                    ],
                    "taxonomy_id": taxid[specie],
                    "species": species[specie],
                    "text_search": specie,
                }

        if specie == "human":
            genomes = add_grch37(genomes, ftp_link)

    ftp.quit()
    return genomes
